// Output .vtk files
// Based on code by M. M. Sussman 11/24/10

// Assumes a finite difference scheme on a geometric mesh
// Assumes number of variables is U.size() / mm.nNode()
// NOT interlaced

#include<map>
#include<vector>
#include<algorithm>
#include<iterator>
#include<sstream>

class VtkFile{
public:
  VtkFile(string name, gmMesh &mm) : fileName(name), pmm(&mm) {}

  ~VtkFile() {if(dump.is_open()) dump.close();}

  void write(feMV::vector &U);

  void write(double tt, feMV::vector &U) // for evolution code
  {
    ostringstream os;

    os << setw(6) << setfill('0') << (int) (tt * 1000) << ".vtk";

    if(dump.is_open()) dump.close();

    dump.open( (fileName + os.str()).c_str() );

    write(U);

    return;
  }

  void resetFileName(string name)
  {
    if(name != fileName)
      {
	if(dump.is_open()) dump.close();
	fileName=name;
	dump.open(fileName.c_str());
      }
  }

  void addScalarName(string name){ scalarNames.push_back(name); }

  void setTitle(string t){ title=t; }

private:
  string fileName, title;
  gmMesh *pmm;
  ofstream dump;

  vector<string> scalarNames;

  void addDefaultScalarName()
  {
    // append integer to word "scalar"
    ostringstream ss;
    ss<<"scalar"<<scalarNames.size();
    scalarNames.push_back(ss.str());
  }

};     // class VtkFile


void VtkFile::write(feMV::vector &U)
{
  int numNodes = (*pmm).nNode();
  vector<int> eNodes(Nchi);
  vector< vector<int> > eltNodes;
  vector<dvector> nodalData(numNodes);

  int numScalars = U.size() / numNodes;

  assert(U.size() == numScalars*numNodes);

  for (gmMesh::geiterator geit = (*pmm).gebegin();
       geit != (*pmm).geend(); geit++)
    {
    gelement ge = *geit;

    for(int i = 0; i < Nchi; i++)
      {
	int ii = ge.nodes[i];

	assert((-1 < ii) && (ii < numNodes));

	nodalData[ii] = ge.xx[i];

	eNodes[i] = ge.nodes[i];
      }

    eltNodes.push_back(eNodes);
    }

  // write the file

  if(!dump.is_open()) dump.open(fileName.c_str());

  // identify ParaView element type

  int vtkEltType; 

  if(Ndim==1 && Nchi==2)      vtkEltType=3;   // line
  else if(Ndim==2 && Nchi==4) vtkEltType=9;   // quadrilateral
  else if(Ndim==2 && Nchi==3) vtkEltType=5;   // triangle
  else if(Ndim==3 && Nchi==4) vtkEltType=10;  // tetrahedron
  else if(Ndim==3 && Nchi==8) vtkEltType=12;  // hexahedron
  else{
    cout << "VtkFile::write(): element type not supported" << endl;
    cerr << "VtkFile::write(): element type not supported" << endl;
    assert(false);
  }

  // local element numbering may differ between ParaView and this
  // hex elements need renumbering,
  // triangles, quads, and tets look OK

  vector<int> renumber(Nchi);

  for(int i=0;i<Nchi;i++) renumber[i]=i;
  
  if (vtkEltType==12) { // renumber hex
    renumber[2]=3;
    renumber[3]=2;
    renumber[6]=7;
    renumber[7]=6;
  }

  dump.precision(8);  // Number of decimal digits

  dump << "# vtk DataFile Version 2.0" << endl;

  if (title.empty()) dump << "Default title" << endl;
  else               dump << title           << endl;

  dump << "ASCII" << endl << endl;  // not writing binary files
  dump << "DATASET UNSTRUCTURED_GRID" << endl;

  dump << "POINTS " << numNodes << " double" << endl;

  for(int node = 0; node < numNodes; node++)
    {
    for(int i = 0; i < 3; i++)
      {
      if(i < Ndim)
	{
	dump.width(15); 
	dump << nodalData[node][i] << " " ;
	}
      else
	dump << "     0.0";
      }
    dump << endl;
    }
    
  dump << "CELLS " << eltNodes.size() << " " 
       << (Nchi+1)*eltNodes.size() << endl;

  for(unsigned int elt = 0; elt < eltNodes.size(); elt++)
    {
    dump << Nchi << " ";

    for(int i = 0; i < Nchi; i++)
      {
      dump.width(9);  // only used to make columns straight
      dump << eltNodes[elt][renumber[i]] << " ";
      }
    dump << endl;
    }

  dump << endl;
 
  dump << "CELL_TYPES " << eltNodes.size() << endl;

  for(unsigned int elt=0; elt < eltNodes.size(); elt++)
    dump << vtkEltType << endl;

  dump << endl;

  dump << "POINT_DATA " << numNodes << endl;

  for(int i=0;i<numScalars;i++)
    {
      if((int) scalarNames.size() <= i) addDefaultScalarName();
    
    dump << "SCALARS " << scalarNames[i] << " double" << endl;
    dump << "LOOKUP_TABLE default" << endl;  // default colors

    for(int node = 0; node < numNodes; node++)
      {
      dump.width(25); 
      dump << U[node + i * (*pmm).nNode()] << endl;
      }
    dump << endl;
  }

  return;
}              // end write()


