#include"triangleMesh.C"

#include"vcell.C"

typedef VoronoiCell<2> vcell;
typedef BoundaryVcell<2> bvcell;

class triangleVmesh : public triangleMesh
{
public:
  triangleVmesh(std::string);        

  triangleVmesh(const triangleVmesh &mm) 
  {
    std::cerr << "triangleVmesh(): don't copy a triangleVmesh\n";
    throw("copying a triangleVmesh");
  }
    
  triangleVmesh() 
  {
    std::cerr << "triangleVmesh(): Default constructor called, nyi\n";
    throw("bad triangleVmesh constructor");
  }

  ~triangleVmesh() {return;}

  int nVedge()  const {return nvedge;}

  void vprint();

  class viterator : public std::vector<svector<2> >::iterator
  {
  public:
    viterator(const std::vector<svector<2> >::iterator &it, triangleVmesh *vp) 
      : std::vector<svector<2> >::iterator(it), vm(vp) {}

    vcell operator*();
    
  private:
    triangleVmesh *vm;
  };

  viterator vbegin() {return viterator(vertices.begin(), this);}
  viterator vend()   {return viterator(vertices.end(),   this);}

  // Boundary itterator

  class bviterator : public std::map<int, pair<int,int> >::iterator
  {
  public:
    bviterator(const std::map<int, pair<int,int> >::iterator &it, 
	       triangleVmesh *vp) 
      : std::map<int, pair<int,int> >::iterator(it), vm(vp) {}

    bvcell operator*();
    
  private:
    triangleVmesh *vm;
  };

  bviterator bvbegin() {return bviterator(vbdy.begin(), this);}
  bviterator bvend()   {return bviterator(vbdy.end(),   this);}


  // bvcell nyi

private:
  int nvedge;

  // vertices[], segments[] and triangles[] inherited from the triangleMesh

  // Convention:
  // Voronoi edges, vedges[], are pairs of distinct integers (i,j)
  //  1) i>=0 and tcenters[i] is one end (circumcenter of triangles[i])
  //  2) If   j>=0 then it an index to tcenters[] as above
  //     Else j< 0 and the mid point of segments[-j-1] is the other end
  
  // Note: Voronoi edges are not orientiented and vnbrs[] not ordered
  
  // tcenters[] circumcenters of the triangles
  // vnbrs[]  vector  of (primal vertices[], dual vedges[]) indices
  // vedges[] indices of (triangle[], triangle[]) or (triangle[], -1-segment[])
  // vbdy[i]  indices of neighboring (vertices[], vertices[]) 

  std::vector<svector<2> > tcenters;     

  std::vector<std::vector<std::pair<int,int> > > vnbrs;

  std::vector<std::pair<int,int> > vedges;  // dual edges

  std::map<int, pair<int,int> > vbdy;       
  std::map<int, pair<svector<2> ,svector<2> > > bnormal; // boundary normals
  std::map<int, int> btype;                              // boundary types

  // For efficientcy compute edge lengths and Voronoi areas

  std::vector<double> varea;                 // Voronoi areas
  std::map<pair<int,int>, double> hlength;   // primal edge lengths
  std::map<pair<int,int>, double> vlength;   // dual edge lengths
  
  void vinput(std::string meshFile);
};

triangleVmesh::triangleVmesh(std::string meshFile)
  : triangleMesh(meshFile)
{
  std::cout << "triangleVmesh(" << meshFile << "): Constructing Voronoi"  
	    << " diagram (reads .v.node, .v.edge)" << std::endl;

  tcenters.resize(nelem);
  vnbrs.resize(nodes);
  varea.resize(nodes, 0.0);

  vinput(meshFile);   // read in the Voronoi diagram data

  return;
}

void triangleVmesh::vinput(std::string meshFile)
{
  int nAttr, nBmkr, dim, index, i0,i1, j0,j1, ntri;
  double xx,yy;

  // Read in the triangle circumcenters (nodes of Voronoi diagram)

  std::ifstream in((meshFile + ".v.node").c_str());

  if( !in.is_open() ) 
    {
      std::cerr << "can't open file:" << meshFile + ".v.node" << std::endl;

      throw("Can't open file");
    }

  in >> ntri >> dim >> nAttr >> nBmkr;

  assert(dim == 2);

  assert(ntri == nelem);

  for(int i = 0; i < ntri; i++)
    {
    in.ignore(256,'\n');

    in >> index;
    
    if(index != i) 
      std::cerr << "triangleVmesh::input(): Warning: data in file "
		<< meshFile + ".v.node" << " may be bad\n";
    
    in >> xx >> yy;
    
    tcenters[i] = svector<2>(xx,yy);
    
    // ignore the rest
    }

  in.close();

  // Construct the lookup table for boundary segments
  // boundary(i,j) = k if and only if semgments[k] = (i,j)

  std::map<std::pair<int,int>, int> boundary;  

  for(int i = 0; i < nbelem; i++) 
    {
      std::pair<int,int> ss(segments[i][0], segments[i][1]);
      
      boundary[ss] = i;   
    }
  
  // Read in dual edges from the .v.edge file

  in.open((meshFile + ".v.edge").c_str());

  if( !in.is_open() ) 
    {
      std::cerr << "can't open file:" << meshFile + ".v.edge" << std::endl;

      throw("Can't open file");
    }

  in >> nvedge;

  assert(nvedge >= 0);

  vedges.resize(nvedge);
    
  for(int i = 0; i < nvedge; i++)
    {
    in.ignore(256,'\n');

    in >> index;

    if(index != i)
      std::cerr << "Warning: data in .v.edge file may be bad" << std::endl;
   
    in >> i0 >> i1;
    
    assert((-1 < i0) && (i0 < nelem));    // first index must be triangle
    assert((-2 < i1) && (i1 < nelem));    // second may be -1

    if(i1 > -1)           // two triangles
      {
      int v0[3], *v1;     // size 3 in case of input error
	
      int *t00 = triangles[i0], *t10 = triangles[i1]; 

      set<int> t0(t00, t00+3);           // will order vertices
      set<int> t1(t10, t10+3);           // for intersection
      
      v1 = set_intersection(t0.begin(),t0.end(), t1.begin(),t1.end(), v0);

      assert(v1 == v0+2);    // two vertices of edge in common

      assert((-1 < v0[0]) && (v0[0] < nodes));
      assert((-1 < v0[1]) && (v0[1] < nodes));

      vnbrs[v0[0]].push_back(std::pair<int,int>(v0[1],i));
      vnbrs[v0[1]].push_back(std::pair<int,int>(v0[0],i));
      
      vedges[i] = std::pair<int,int>(i0,i1);

      double vh = (tcenters[i0] - tcenters[i1]).norm();

      vlength[vedges[i]] = vh;

      double hh = (vertices[v0[0]] - vertices[v0[1]]).norm();

      hlength[std::pair<int,int>(v0[0],v0[1])] = hh;
      hlength[std::pair<int,int>(v0[1],v0[0])] = hh;

      varea[v0[0]] += 0.25 * hh*vh;     
      varea[v0[1]] += 0.25 * hh*vh;     
      }
    else            // i1 = -1 indicates a boundary
      {
      in >> xx >> yy;  

      svector<2> nn(xx,yy);      // should be perpendicular to an edge
      
      double modn = nn.norm(); 
      
      assert(modn > 1.0e-8);
      
      nn = (1.0/modn) * nn;

      int j;

      for(j = 0; j < 3; j++)     // locate the edge of triangle[i0]
	{
	j0 = triangles[i0][j];
	j1 = triangles[i0][(j+2)%3];

	svector<2> e01 = vertices[j1]-vertices[j0];
	
	if(fabs(nn.dot(e01)) < 1.0e-12) break;
	}
	
      assert(j < 3);    // should find the edge
      
      std::pair<int,int> s0(j0,j1), s1(j1,j0);
	
      if(     boundary.find(s0) != boundary.end()) i1 = boundary[s0];
      else if(boundary.find(s1) != boundary.end()) i1 = boundary[s1];
      else 
	throw("triangleVmesh::vinput() boundary segment not found");

      int v0 = segments[i1][0], v1 = segments[i1][1];
	
      vnbrs[v0].push_back(std::pair<int,int>(v1,i));
      vnbrs[v1].push_back(std::pair<int,int>(v0,i));

      vedges[i] = pair<int,int>(i0,-1-i1);  

      svector<2> smid = 0.5*(vertices[v0] + vertices[v1]);

      double vh = (tcenters[i0] - smid).norm();

      vlength[vedges[i]] = vh;

      double hh = (vertices[v0] - vertices[v1]).norm();

      hlength[std::pair<int,int>(v0,v1)] = hh;
      hlength[std::pair<int,int>(v1,v0)] = hh;

      varea[v0] += 0.25 * hh*vh;     
      varea[v1] += 0.25 * hh*vh;     

      // Boundary cells

      svector<2> zz(0,0);

      if(vbdy.find(v0) == vbdy.end()) 
	{
	  btype[v0] = segments[i1][2];

	  vbdy[v0] = std::pair<int,int>(v1,-1);
	  bnormal[v0] = std::pair<svector<2>, svector<2> > (nn,zz);
	}
      else                            
	{
	  vbdy[v0].second = v1;
	  bnormal[v0].second = nn;
	}

      if(vbdy.find(v1) == vbdy.end()) 
	{
	  btype[v1] = segments[i1][2];

	  vbdy[v1] = std::pair<int,int>(v0,-1);
	  bnormal[v1] = std::pair<svector<2>, svector<2> >(nn,zz);
	}
      else 
	{
	  vbdy[v1].second = v0;
	  bnormal[v1].second = nn;
	}
      }
    }
  
  in.close();

  return;
}

vcell triangleVmesh::viterator::operator*() 
{
  vcell cc;

  int idx = std::distance(vm->vertices.begin(), 
			  (std::vector<svector<2> >::iterator) *this);

  cc.nNbr = vm->vnbrs[idx].size();

  cc.node = idx;
  cc.xx = vm->vertices[idx];
  cc.volume = vm->varea[idx];

  cc.neighbors.resize(cc.nNbr);
  cc.xNeighbors.resize(cc.nNbr);
  cc.hEdges.resize(cc.nNbr);
  cc.aFaces.resize(cc.nNbr);

  for(int i = 0; i < cc.nNbr; i++)
    {
    cc.neighbors[i] = vm->vnbrs[idx][i].first;
    cc.xNeighbors[i] = vm->vertices[vm->vnbrs[idx][i].first];

    cc.hEdges[i] = vm->hlength[std::pair<int,int>(idx,cc.neighbors[i])];
    cc.aFaces[i] = vm->vlength[vm->vedges[vm->vnbrs[idx][i].second]];
    }    
  
  return cc;
}

bvcell triangleVmesh::bviterator::operator*() 
{
  bvcell cc;

  pair<int, pair<int,int> > ijj 
    = std::map<int, pair<int,int> >::iterator::operator*();

  cc.nNbr = 2;

  cc.node = ijj.first;
  cc.type = vm->btype[ijj.first];
  cc.xx = vm->vertices[ijj.first];

  cc.neighbors.resize(2);
  cc.xNeighbors.resize(2);
  cc.hEdges.resize(2);
  cc.aFaces.resize(2);
  cc.normals.resize(2);

  pair<int,int> jj = ijj.second;

  cc.neighbors[0] = jj.first;  cc.xNeighbors[0] = vm->vertices[jj.first];
  cc.neighbors[1] = jj.second; cc.xNeighbors[1] = vm->vertices[jj.second];

  cc.hEdges[0] = (cc.xx - cc.xNeighbors[0]).norm();
  cc.hEdges[1] = (cc.xx - cc.xNeighbors[1]).norm();

  cc.aFaces[0] = 0.5 * cc.hEdges[0];
  cc.aFaces[1] = 0.5 * cc.hEdges[1];
  
  cc.area = cc.aFaces[0] + cc.aFaces[1];

  cc.normals[0] = vm->bnormal[ijj.first].first;
  cc.normals[1] = vm->bnormal[ijj.first].second;

  return cc;
}

void triangleVmesh::vprint()
{
  std::ofstream dump("outputVmesh");

  dump.setf(std::ios::fixed, std::ios::floatfield);

  dump << "triangleVmesh::print(): nodes = " << nodes << endl;
  dump << "triangleVmesh::print(): vnbrs.size() = " << vnbrs.size() << endl;
  dump << "triangleVmesh::print(): vedges.size() = " << vedges.size() << endl;

  for(viterator vit = vbegin(); vit != vend(); vit++)
    {
      vcell cc = *vit;

      dump << "Cell " << cc.node 
	   << ", type = " << cc.type
	// << ", No. neighbors = " << cc.nNbr
	   << ", Area = " << cc.volume
	   << ", Coordinates = " << cc.xx
	   << endl;

      for(int i = 0; i < cc.nNbr; i++)
	{
	  dump << "\tneighbor[" << i << "] = " << cc.neighbors[i]
	   << ", edge length = " << cc.hEdges[i]
	   << ", Voronoi area = " << cc.aFaces[i]
	   << ", coordinates = " << cc.xNeighbors[i]
	   << endl;
	}
    }

  dump << endl
       << "triangleVmesh::print(): vbdy.size() = " 
       << vbdy.size() << endl;

  for(bviterator bit = bvbegin(); bit != bvend(); bit++)
    {
      bvcell cc = *bit;

      assert(cc.nNbr == 2);

      dump << "Cell " << cc.node 
	   << ", type = " << cc.type
	// << ", No. neighbors = " << cc.nNbr
	   << ", Area = " << cc.area
	   << ", Coordinates = " << cc.xx
	   << endl;

      dump << "    neighbor[0] = " << cc.neighbors[0]
	   << ", edge length = " << cc.hEdges[0]
	   << ", boundary length = " << cc.aFaces[0]
	   << endl
	   << "\tcoordinates = " << cc.xNeighbors[0]
	   << ", normal = " << cc.normals[0]
	   << endl;

      dump << "    neighbor[1] = " << cc.neighbors[1]
	   << ", edge length = " << cc.hEdges[1]
	   << ", boundary length = " << cc.aFaces[1]
	   << endl
	   << "\tcoordinates = " << cc.xNeighbors[1]
	   << " normal = " << cc.normals[1]
	   << endl;
    }
}
